"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SagaBuilder = exports.Saga = exports.SagaStatus = exports.SagaCompensationError = exports.SagaInvocationError = void 0;
class SagaInvocationError extends Error {
    constructor(originalError, step) {
        super(originalError.message);
        this.originalError = originalError;
        this.step = step;
        Object.setPrototypeOf(this, SagaInvocationError.prototype);
        this.stack = originalError.stack;
    }
}
exports.SagaInvocationError = SagaInvocationError;
class SagaCompensationError extends Error {
    constructor(originalError, step) {
        super(originalError.message);
        this.originalError = originalError;
        this.step = step;
        Object.setPrototypeOf(this, SagaCompensationError.prototype);
        this.stack = originalError.stack;
    }
}
exports.SagaCompensationError = SagaCompensationError;
class Step {
    constructor(name, invocation) {
        this.name = name;
        this.invocation = invocation;
    }
    async invoke(params) {
        return this.invocation(params);
    }
    async compensate(params) {
        if (this.compensation)
            return this.compensation(params);
    }
    bind(ctx) {
        this.invocation = this.invocation.bind(ctx);
        if (this.compensation)
            this.compensation = this.compensation.bind(ctx);
        return this;
    }
}
class SagaFlow {
    constructor(steps, returnFn) {
        this.steps = steps;
        this.returnFn = returnFn;
        this.compensationSteps = [];
    }
    get current() {
        return this.__current;
    }
    async invoke(params) {
        for (const step of this.steps) {
            this.__current = step;
            await step.invoke(params);
            this.compensationSteps.push(step);
        }
        return this.returnFn(params);
    }
    async compensate(params) {
        for (const step of this.compensationSteps.reverse()) {
            this.__current = step;
            await step.compensate(params);
        }
    }
    bind(ctx) {
        this.steps.forEach((step) => step.bind(ctx));
        return this;
    }
}
var SagaStatus;
(function (SagaStatus) {
    SagaStatus["New"] = "New";
    SagaStatus["InProgress"] = "In progress";
    SagaStatus["InCompensation"] = "In compensation";
    SagaStatus["Complete"] = "Complete";
    SagaStatus["CompensationComplete"] = "Compensation complete";
    SagaStatus["CompensationError"] = "Compensation error";
})(SagaStatus = exports.SagaStatus || (exports.SagaStatus = {}));
class Saga {
    constructor(sagaFlow) {
        this.sagaFlow = sagaFlow;
        this.__status = SagaStatus.New;
    }
    get status() {
        return this.__status;
    }
    async execute(params) {
        this.__status = SagaStatus.InProgress;
        try {
            const result = await this.sagaFlow.invoke(params);
            this.__status = SagaStatus.Complete;
            return result;
        }
        catch (e) {
            this.__status = SagaStatus.InCompensation;
            await this.runCompensationFlow(params);
            throw new SagaInvocationError(wrapIfNotError(e), this.sagaFlow.current.name);
        }
    }
    bind(ctx) {
        this.sagaFlow.bind(ctx);
        return this;
    }
    async runCompensationFlow(params) {
        try {
            await this.sagaFlow.compensate(params);
            this.__status = SagaStatus.CompensationComplete;
        }
        catch (e) {
            this.__status = SagaStatus.CompensationError;
            throw new SagaCompensationError(wrapIfNotError(e), this.sagaFlow.current.name);
        }
    }
}
exports.Saga = Saga;
function wrapIfNotError(x) {
    if (!(x instanceof Error)) {
        return new Error(String(x));
    }
    return x;
}
class SagaBuilder {
    constructor() {
        this.steps = [];
        this.returnFn = (() => void 0);
    }
    step(name = '') {
        return {
            invoke: (method) => {
                const step = new Step(name || method.name || `step${this.steps.length}`, method);
                this.steps.push(step);
                const nextOptions = {
                    step: (name = '') => this.step(name),
                    return: (method) => this.return(method),
                    build: () => this.build(),
                };
                return {
                    ...nextOptions,
                    withCompensation: (method) => {
                        step.compensation = method;
                        return nextOptions;
                    },
                };
            },
        };
    }
    return(method) {
        this.returnFn = method;
        return { build: () => this.build() };
    }
    build() {
        return new Saga(new SagaFlow(this.steps, this.returnFn));
    }
}
exports.SagaBuilder = SagaBuilder;
//# sourceMappingURL=core.js.map